﻿------------------------------------------------------------------------------
--	GridStatusPriestAoe by Slaren
--    Thanks to Duugu for GridStatusCloseUp idea and original code
------------------------------------------------------------------------------

local GridStatus = Grid:GetModule("GridStatus")

GridStatusPriestAoe = GridStatus:NewModule("GridStatusPriestAoe")
GridStatusPriestAoe.menuName = "Priest Aoe Healing"

-- constants
local TALENTTREE_DISCIPLINE = 1
local TALENTTREE_HOLY = 2

local SPELLID_COH = 34861
local SPELLID_POH = 596
local SPELLID_CASC = 121135
local SPELLID_TOF = 123254

local GLYPH_COH = 55675

local ICON_COH = "Interface\\Icons\\Spell_Holy_CircleOfRenewal"
local ICON_POH = "Interface\\Icons\\Spell_Holy_PrayerOfHealing02"
local ICON_HN = "Interface\\Icons\\Spell_Holy_HolyNova"
local ICON_CASC = "Interface\\Icons\\ability_priest_cascade"

local COH_BASE =  0
local POH_BASE =  0
local HN_BASE =   0
local CASC_BASE = 0

local SPELLMOD_COH = 2.21666
local SPELLMOD_POH = 2.21664
local SPELLMOD_HN =  1.125
local SPELLMOD_CASC_MIN = 0.216
local SPELLMOD_CASC_MAX = 0.54

local CASC_JUMP_RANGE_SQ = 40 ^ 2

local PARTY_STATE_GROUP_COUNT = {
	["solo"] = 1,
	["party"] = 1,
	["raid_10"] = 2,
	["raid_25"] = 5,
	["raid_flex"] = 8,
	["bg"] = 8,
	["arena"] = 1
}

-- upvalues
local GetSpellBonusHealing = GetSpellBonusHealing
local GetSpellCooldown = GetSpellCooldown
local GetSpellCritChance = GetSpellCritChance
local GetMultistrike = GetMultistrike
local GetMasteryEffect = GetMasteryEffect
local GetCombatRatingBonus = GetCombatRatingBonus
local GetVersatilityBonus = GetVersatilityBonus
local CR_VERSATILITY_DAMAGE_DONE = CR_VERSATILITY_DAMAGE_DONE
local GetTime = GetTime
local GridFrame = Grid:GetModule("GridFrame")
local GridRoster = Grid:GetModule("GridRoster")
local GridStatusPriestAoe = GridStatusPriestAoe
local SetMapToCurrentZone = SetMapToCurrentZone
local UnitBuff = UnitBuff
local UnitGUID = UnitGUID
local UnitGetIncomingHeals = UnitGetIncomingHeals
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitInRange = UnitInRange
local UnitIsCharmed = UnitIsCharmed
local UnitIsConnected = UnitIsConnected
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsEnemy = UnitIsEnemy
local UnitIsUnit = UnitIsUnit
local UnitIsVisible = UnitIsVisible
local UnitPosition = UnitPosition
local math_floor = math.floor
local math_max = math.max
local math_min = math.min
local math_sqrt = math.sqrt
local pairs = pairs
local select = select
local tinsert = table.insert
local tostring = tostring
local tsort = table.sort
local wipe = wipe

-- local data
local settings
local settings_coh
local settings_poh
local settings_hn
local settings_casc
local refresh_state_player_x
local refresh_state_player_y
local refresh_state_player_data = {}
local num_enabled = 0
local update_frame = CreateFrame("Frame")
local coh_prev_best_guid
local coh_prev_best_heal
local casc_prev_best_guid
local casc_prev_best_heal
local poh_prev_best
local coh_targets
local update_timer = 0
local coh_testrange_sq
local poh_testrange_sq
local hn_testrange_sq
local num_groups
local test_of_faith_mul
local roster
local player_group
local player_guid
local have_casc
local spellname_tof
local is_discipline
local base_coef
local spell_power
local bonus_coef


-- Options
GridStatusPriestAoe.defaultDB = {
	debug = false,
	cycle_time = 1 / 3,
	first_groups_only = false,

	alert_priestaoe_coh = {
		-- coh
		enable = true,
		color = { r = 1, g = 1, b = 0, a = 1 },
		priority = 99,
		range = true,

		enable_if_have = true,
		threshold_targets = 4,
		hidecd = true,

		emphasize = true,
		emphasize_color = { r = 1, g = 1, b = 0 },
		emphasize_priority = 51,
	},
	alert_priestaoe_poh = {
		-- poh
		enable = true,
		color = { r = 1, g = 1, b = 1, a = 1 },
		priority = 97,
		range = true,

		always_best = false,
		threshold_targets = 4,
		incoming_heals = false,
		
		emphasize = false,
		emphasize_color = { r = 1, g = 1, b = 1 },
		emphasize_priority = 49,
	},
	alert_priestaoe_casc = {
		-- cascade
		enable = false,
		color = { r = 0, g = 0, b = 1, a = 1 },
		priority = 98,
		range = true,

		enable_if_have = true,
		threshold_targets = 15,
		hidecd = true,

		emphasize = true,
		emphasize_color = { r = 0, g = 0, b = 1 },
		emphasize_priority = 50,
	},
	alert_priestaoe_hn = {
		-- holy nova
		enable = false,
		color = { r = 0, g = 1, b = 0, a = 1 },
		priority = 96,

		threshold_targets = 4,
		emphasize = false,
	},
}

local coh_options = {
	["coh_threshold"] = {
		order = 100,
		type = "range",
		name = "Heal threshold",
		desc = "Threshold for heal amount",
		softMax = 6,
		min = 0,
		step = 0.01,
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_coh.threshold_targets end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_coh.threshold_targets = v end,
	},
	["coh_enable_if_have"] = {
		order = 101,
		type = "toggle",
		name = "Show only while having the talent",
		desc = "Enable this if you want to hide this status while using a talent spec that does not include Circle of Healing",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_coh.enable_if_have end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_coh.enable_if_have = v; GridStatusPriestAoe:UpdateTalents(); end,
	},
	["coh_hidecd"] = {
		order = 102,
		type = "toggle",
		name = "Hide while on cooldown",
		desc = "Hide status while Circle of Healing is on cooldown",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_coh.hidecd end,
		set = function (_, arg) GridStatusPriestAoe.db.profile.alert_priestaoe_coh.hidecd = arg end,
	}
}

local poh_options = {
	["poh_threshold"] = {
		order = 100,
		type = "range",
		name = "Heal threshold",
		desc = "Threshold for heal amount",
		softMax = 5,
		min = 0,
		step = 0.01,
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.threshold_targets end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_poh.threshold_targets = v end,
	},
	["poh_enablebest"] = {
		order = 101,
		type = "toggle",
		name = "Always show best target in each group",
		desc = "Shows best target to cast Prayer of Healing on in each group, even if the heal amount is below the threshold (minimum 3 targets)",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.always_best end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_poh.always_best = v end,
	},
	--[[
	["poh_hideglyph"] = {
		order = 102,
		type = "toggle",
		name = "Ignore players affected by your Glyph of Prayer of Healing",
		desc = "Ignores players that have your Glyph of Prayer of Healing buff (modified by your Prayer of Healing cast time)",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.hide_glyph end,
		set = function (_, arg) GridStatusPriestAoe.db.profile.alert_priestaoe_poh.hide_glyph = arg end,
	},
	]]
	["poh_incomingheals"] = {
		order = 103,
		type = "toggle",
		name = "Subtract incoming heals",
		desc = "Subtracts incoming heals from the health deficit before making the calculations (your own heals are excluded)",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.incoming_heals end,
		set = function (_, arg) GridStatusPriestAoe.db.profile.alert_priestaoe_poh.incoming_heals = arg end,
	},
}

local casc_options = {
	["range"] = false,
	["casc_threshold"] = {
		order = 100,
		type = "range",
		name = "Heal threshold",
		desc = "Threshold for heal amount",
		min = 0,
		softMax = 31,
		step = 0.01,
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_casc.threshold_targets end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_casc.threshold_targets = v end,
	},
	["casc_enable_if_have"] = {
		order = 101,
		type = "toggle",
		name = "Show only while having the talent",
		desc = "Enable this if you want to hide this status while using a talent spec that does not include Cascade",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_casc.enable_if_have end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_casc.enable_if_have = v; GridStatusPriestAoe:UpdateTalents(); end,
	},
	["casc_hidecd"] = {
		order = 102,
		type = "toggle",
		name = "Hide while on cooldown",
		desc = "Hide status while Cascade is on cooldown",
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_casc.hidecd end,
		set = function (_, arg) GridStatusPriestAoe.db.profile.alert_priestaoe_casc.hidecd = arg end,
	}	
}

local hn_options = {
	["range"] = false,
	["hn_threshold"] = {
		order = 100,
		type = "range",
		name = "Heal threshold",
		desc = "Threshold for heal amount",
		min = 1,
		softMax = 100000,
		step = 1,
		width = "full",
		get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_hn.threshold end,
		set = function (_, v) GridStatusPriestAoe.db.profile.alert_priestaoe_hn.threshold = v end,
	}
}

GridStatusPriestAoe.extraOptions = {
	["frequency"] = {
		order = 100,
		type = "range",
		name = "Refresh frequency",
		desc = "Number of status refreshes per second",
		width = "full",
		min = 1,
		softMax = 30,
		step = 1,
		get = function () return 1 / GridStatusPriestAoe.db.profile.cycle_time end,
		set = function (_, v) GridStatusPriestAoe.db.profile.cycle_time = 1 / v end,
	},
	["first_groups_only"] = {
		order = 101,
		type = "toggle",
		name = "Check only first groups",
		desc = "Depending on the raid size, checks only the first groups to improve performance",
		get = function () return GridStatusPriestAoe.db.profile.first_groups_only end,
		set = function (_, v) GridStatusPriestAoe.db.profile.first_groups_only = v; GridStatusPriestAoe:UpdateNumGroups() end,
	},
}

-- Emphasize
local Emphasize = GridStatus:GetModule("GridStatusEmphasize", true)

local function print(...)
	DEFAULT_CHAT_FRAME:AddMessage("|cff33ff99GridStatusPriestAoe|r: " .. table.concat({...}, " "))
end

if Emphasize and not Emphasize.EmphasizeUnit then
	print("Your version of GridStatusEmphasize is not compatible with this version of GridStatusPriestAoe. Please update to the most recent version.")
	Emphasize = nil
end

if Emphasize then
	local PriestAoe_Emphasize = Emphasize:NewModule("PriestAoe_Emphasize")
	local emphasize_options = {
		["coh"] = {
			type = "group",
			inline = true,
			name = "Circle of Healing",
			args = {
				["emphasize"] = {
					order = 1,
					type = "toggle",
					width = "full",
					name = "Emphasize Circle of Healing",
					get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize end,
					set = function (_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize = v
						GridStatusPriestAoe:ClearCohStatus()
					end,
				},
				["color"] = {
					order = 2,
					type = "color",
					hasAlpha = false,
					name = "Color",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize end,
					get = function ()
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize_color
						return color.r, color.g, color.b
					end,
					set = function (_, r, g, b)
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize_color
						color.r = r
						color.g = g
						color.b = b
					end,
				},
				["priority"] = {
					order = 3,
					type = "range",
					max = 99,
					min = 0,
					step = 1,
					name = "Priority",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize end,
					get = function() return GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize_priority end,
					set = function(_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_coh.emphasize_priority = v
						GridStatusPriestAoe:ClearCohStatus()
					end,
				},
			},
		},
		["poh"] = {
			type = "group",
			inline = true,
			name = "Prayer of Healing",
			args = {
				["emphasize"] = {
					order = 1,
					type = "toggle",
					width = "full",
					name = "Emphasize Prayer of Healing",
					get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize end,
					set = function (_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize = v
						GridStatusPriestAoe:ClearPohEmphasizeStatus()
					end,
				},
				["color"] = {
					order = 2,
					type = "color",
					hasAlpha = false,
					name = "Color",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize end,
					get = function ()
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize_color
						return color.r, color.g, color.b
					end,
					set = function (_, r, g, b)
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize_color
						color.r = r
						color.g = g
						color.b = b
					end,
				},
				["priority"] = {
					order = 3,
					type = "range",
					max = 99,
					min = 0,
					step = 1,
					name = "Priority",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize end,
					get = function() return GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize_priority end,
					set = function(_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_poh.emphasize_priority = v
						GridStatusPriestAoe:ClearPohEmphasizeStatus()
					end,
				},				
			}
		},
		-- Cascade
		["casc"] = {
			type = "group",
			inline = true,
			name = "Cascade",
			args = {
				["emphasize"] = {
					order = 1,
					type = "toggle",
					width = "full",
					name = "Emphasize Cascade",
					get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize end,
					set = function (_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize = v
						GridStatusPriestAoe:ClearCascStatus()
					end,
				},
				["color"] = {
					order = 2,
					type = "color",
					hasAlpha = false,
					name = "Color",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize end,
					get = function ()
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize_color
						return color.r, color.g, color.b
					end,
					set = function (_, r, g, b)
						local color = GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize_color
						color.r = r
						color.g = g
						color.b = b
					end,
				},
				["priority"] = {
					order = 3,
					type = "range",
					max = 99,
					min = 0,
					step = 1,
					name = "Priority",
					disabled = function() return not GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize end,
					get = function() return GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize_priority end,
					set = function(_, v)
						GridStatusPriestAoe.db.profile.alert_priestaoe_casc.emphasize_priority = v
						GridStatusPriestAoe:ClearCascStatus()
					end,
				},
			},
		},		
		--[[ Holy Nova
		["hn_emphasize"] = {
			order = 3,
			type = "toggle",
			name = "Holy Nova",
			desc = "Emphasize Holy Nova",
			get = function () return GridStatusPriestAoe.db.profile.alert_priestaoe_hn.emphasize end,
			set = function (_, v)
				GridStatusPriestAoe.db.profile.alert_priestaoe_hn.emphasize = v
				if not v then Emphasize:DeemphasizeAllUnits("priestaoe_hn") end
			end,
		},
		]]
	}
	function PriestAoe_Emphasize:OnInitialize()
		self.super.OnInitialize(self)

		self:RegisterOptions("Priest Aoe Healing", emphasize_options)
	end
end

function GridStatusPriestAoe:UpdateCoefs()
	base_coef = 1

	if select(2, UnitClass("player")) == "PRIEST" then
		if GetSpecialization() == TALENTTREE_HOLY then
			base_coef = 1.3 -- grace
		elseif GetSpecialization() == TALENTTREE_DISCIPLINE then
			base_coef = 1.1
		end
	end

	test_of_faith_mul = 1
end

local function UpdateBonusCoefs()
	local spellcrit = GetSpellCritChance(2)
	local multistrike = GetMultistrike()
	local mastery = GetMasteryEffect()
	local versatility = GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_DONE) + GetVersatilityBonus(CR_VERSATILITY_DAMAGE_DONE)

	local crit_coef = 1 + spellcrit / 200
	local ms_coef = 1 + multistrike / 100 * 0.6
	local mastery_coef = 1 + mastery / 200
	local versatility_coef = 1 + versatility / 100

	if is_discipline then
		bonus_coef = base_coef * mastery_coef * versatility_coef
	else
		bonus_coef = base_coef * versatility_coef
	end

	bonus_coef = bonus_coef * crit_coef * ms_coef

	-- Twist of Fate
	if UnitBuff("player", spellname_tof) then
		bonus_coef = bonus_coef * 1.15
	end

	spell_power = GetSpellBonusHealing()
end

local function GetSpellAvg(base, coef_madd, coef_m, spellmod)
	-- note: poh crit +10%
	local coef = (1 + coef_madd) * coef_m * bonus_coef

	local total = (base + spell_power * spellmod) * coef
	
	return math_floor(total + 0.5)
end

local function GetCohAvg()
	return GetSpellAvg(COH_BASE, 0, 1, SPELLMOD_COH)
end

local function GetPohAvg()
	return GetSpellAvg(POH_BASE, 0, 1, SPELLMOD_POH)
end

local function GetHnAvg()
	return GetSpellAvg(HN_BASE, 0, 1, SPELLMOD_HN)
end

local function GetCascAvg(dist)
	local dist_mod = math_min(dist, 30) / 30
	local spellmod = SPELLMOD_CASC_MIN + (SPELLMOD_CASC_MAX - SPELLMOD_CASC_MIN) * dist_mod
	return GetSpellAvg(CASC_BASE, 0, 1, spellmod)
end

local function IsCohReady()
	local start, duration = GetSpellCooldown(SPELLID_COH)

	-- ignore global cooldown, show if ready or less than 0.5 secs of cooldown remaining
	if start == 0 or duration <= 1.5 or ((duration - (GetTime() - start)) <= 0.5) then
		return true
	end

	return false
end

local function IsCascReady()
	local start, duration = GetSpellCooldown(SPELLID_CASC)

	-- ignore global cooldown, show if ready or less than 0.5 secs of cooldown remaining
	if start == 0 or duration <= 1.5 or ((duration - (GetTime() - start)) <= 0.5) then
		return true
	end

	return false
end

function GridStatusPriestAoe:OnInitialize()
	self.super.OnInitialize(self)
	-- self.core.core:AddModuleDebugMenu(self)

	self:RegisterStatus("alert_priestaoe_coh", string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_COH) .. " Circle of Healing", coh_options)
	self:RegisterStatus("alert_priestaoe_poh", string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_POH) .. " Prayer of Healing", poh_options)
	self:RegisterStatus("alert_priestaoe_casc", string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_CASC) .. " Cascade", casc_options)
	--[[ Holy Nova
	self:RegisterStatus("alert_priestaoe_hn", "Priest Aoe - Holy Nova", hn_options)
	]]
	
	GridStatus.registryDescriptions["alert_priestaoe_coh"] = string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_COH) .. " Priest Aoe - Circle of Healing"
	GridStatus.registryDescriptions["alert_priestaoe_poh"] = string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_POH) .. " Priest Aoe - Prayer of Healing"
	GridStatus.registryDescriptions["alert_priestaoe_casc"] = string.format("|T%s:0:0:0:0:64:64:4:60:4:60|t", ICON_CASC) .. " Priest Aoe - Cascade"

	settings = self.db.profile
	settings_coh = self.db.profile.alert_priestaoe_coh
	settings_poh = self.db.profile.alert_priestaoe_poh
	settings_hn = self.db.profile.alert_priestaoe_hn
	settings_casc = self.db.profile.alert_priestaoe_casc
end

function GridStatusPriestAoe:OnStatusEnable(status)
	if status == "alert_priestaoe_coh" or status == "alert_priestaoe_poh" or status == "alert_priestaoe_hn" or status == "alert_priestaoe_casc" then
		num_enabled = num_enabled + 1
		if num_enabled == 1 then
			self:UpdateRoster()
			self:UpdateNumGroups()

			self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateTalents")
			self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "UpdateCoefs")
			self:RegisterEvent("GROUP_ROSTER_UPDATE", "UpdateRoster")
			self:RegisterMessage("Grid_PartyTransition", "UpdateNumGroups")

			-- check talents the first frame, then change OnUpdate script to the real function
			update_frame:SetScript("OnUpdate",
				function()
					spellname_tof = GetSpellInfo(SPELLID_TOF)
					
					player_guid = UnitGUID("player")
					
					self:UpdateTalents()
					
					update_frame:SetScript("OnUpdate", function(_, elapsed) return self:OnUpdate(elapsed) end)
				end)
		end
	end
end

function GridStatusPriestAoe:OnStatusDisable(status)
	if num_enabled > 0 then
		if status == "alert_priestaoe_coh" then
			self.core:SendStatusLostAllUnits("alert_priestaoe_coh")
			if Emphasize then Emphasize:DeemphasizeAllUnits("priestaoe_coh") end
			num_enabled = num_enabled - 1
		elseif status == "alert_priestaoe_poh" then
			self.core:SendStatusLostAllUnits("alert_priestaoe_poh")
			if Emphasize then Emphasize:DeemphasizeAllUnits("priestaoe_poh") end
			num_enabled = num_enabled - 1
		elseif status == "alert_priestaoe_hn" then
			self.core:SendStatusLostAllUnits("alert_priestaoe_hn")
			if Emphasize then Emphasize:DeemphasizeAllUnits("priestaoe_hn") end
			num_enabled = num_enabled - 1
		elseif status == "alert_priestaoe_casc" then
			self.core:SendStatusLostAllUnits("alert_priestaoe_casc")
			if Emphasize then Emphasize:DeemphasizeAllUnits("priestaoe_casc") end
			num_enabled = num_enabled - 1
		end

		if num_enabled == 0 then
			update_frame:SetScript("OnUpdate", nil)
			self:UnregisterEvent("PLAYER_TALENT_UPDATE")
			self:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED")
			self:UnregisterEvent("GROUP_ROSTER_UPDATE")
			self:UnregisterMessage("Grid_PartyTransition")
		end
	end
end

function GridStatusPriestAoe:UpdateRoster()
	self:ClearAllStatus()

	roster = { [1] = {}, [2] = {}, [3] = {}, [4] = {}, [5] = {}, [6] = {}, [7] = {}, [8] = {} }

	if not IsInRaid() then
		tinsert(roster[1], "player")
		player_group = 1
		for i = 1, 4 do
			local unitid = "party" .. i
			if UnitExists(unitid) then
				tinsert(roster[1], unitid)
			end
		end
	else
		for i = 1, GetNumGroupMembers() do
			local name, _, group = GetRaidRosterInfo(i)
			local unitid = "raid" .. i
			if name and UnitExists(unitid) then
				tinsert(roster[group], unitid)
				if UnitIsUnit("player", unitid) then
					player_group = group
				end
			end
		end
	end

	self:Debug("roster updated:", #roster[1], #roster[2], #roster[3], #roster[4], #roster[5], #roster[6], #roster[7], #roster[8])
end

function GridStatusPriestAoe:UpdateTalents()
	-- check coh
	if (select(2, UnitClass("player")) == "PRIEST") and (GetSpecialization() == TALENTTREE_HOLY) then
		-- check if player has coh glyph
		if self:HasGlyph(GLYPH_COH) then
			coh_targets = 6
		else
			coh_targets = 5
		end
	else
		coh_targets = nil
	end

	is_discipline = (select(2, UnitClass("player")) == "PRIEST") and (GetSpecialization() == TALENTTREE_DISCIPLINE)

	-- check cascade
	if (select(2, UnitClass("player")) == "PRIEST") and (GetSpecialization() == TALENTTREE_HOLY or GetSpecialization() == TALENTTREE_DISCIPLINE) then
		local talentID, name, iconTexture, selected, available = GetTalentInfo(6, 1, GetActiveSpecGroup())
		have_casc = selected
	else
		have_casc = nil
	end

	-- lost coh? clear status
	if settings_coh.enable and settings_coh.enable_if_have and not coh_targets then
		self:ClearCohStatus()
	end

	if settings_casc.enable and settings_casc.enable_if_have and not have_casc then
		self:ClearCascStatus()
	end

	-- update ranges
	coh_testrange_sq = 30 ^ 2
	poh_testrange_sq = 30 ^ 2
	hn_testrange_sq = 10 ^ 2

	-- update spell coefs
	self:UpdateCoefs()

	self:Debug("coh_targets:", coh_targets)
end

function GridStatusPriestAoe:UpdateNumGroups()
	self:ClearAllStatus()
	num_groups = (settings.first_groups_only and PARTY_STATE_GROUP_COUNT[GridRoster:GetPartyState()]) or 8

	self:Debug("num_groups:", num_groups)
end

function GridStatusPriestAoe:ClearCohStatus()
	if coh_prev_best_guid then
		self.core:SendStatusLost(coh_prev_best_guid, "alert_priestaoe_coh")
		if Emphasize then Emphasize:DeemphasizeUnit(coh_prev_best_guid, "priestaoe_coh") end
		coh_prev_best_guid = nil
		coh_prev_best_heal = -1
	end
end

function GridStatusPriestAoe:ClearCascStatus()
	if casc_prev_best_guid then
		self.core:SendStatusLost(casc_prev_best_guid, "alert_priestaoe_casc")
		if Emphasize then Emphasize:DeemphasizeUnit(casc_prev_best_guid, "priestaoe_casc") end
		casc_prev_best_guid = nil
		casc_prev_best_heal = -1
	end
end

function GridStatusPriestAoe:ClearPohGridStatus()
	self.core:SendStatusLostAllUnits("alert_priestaoe_poh")
end

function GridStatusPriestAoe:ClearPohEmphasizeStatus()
	if Emphasize then Emphasize:DeemphasizeUnit(poh_prev_best, "priestaoe_poh") end
	poh_prev_best = nil
end

function GridStatusPriestAoe:ClearPohStatus()
	self:ClearPohGridStatus()
	self:ClearPohEmphasizeStatus()
end

function GridStatusPriestAoe:ClearAllStatus()
	self:ClearCohStatus()
	self:ClearPohStatus()
	self:ClearCascStatus()
end

--[[
function GridStatusPriestAoe:ClearHnStatus()
	self.core:SendStatusLost(player_guid, "alert_priestaoe_hn")
	if Emphasize then Emphasize:DeemphasizeUnit(player_guid, "priestaoe_hn") end
end
]]

-- debug
local update_init_time = nil
local update_time = nil
local update_this_start_time = 0
local debug_frame = nil
local debug_fs = nil
local GetTime = GetTime

local function start_timing()
	if not debug_frame then
		debug_frame = CreateFrame("Frame", "gridstatuspriestaoe_performance_monitor", UIParent)
		debug_fs = debug_frame:CreateFontString(nil, "OVERLAY")
		debug_fs:SetFont(STANDARD_TEXT_FONT, 18,"THICKOUTLINE")
		debug_fs:SetPoint("LEFT", UIParent, "LEFT")
		debug_fs:SetText("GridStatusPriestAoe performance monitor")
	end
	local ctime = GetTime()
	update_this_start_time = ctime
	if update_init_time == nil then
		update_init_time = ctime
		update_time = 0
	end
end

local function end_timing()
	local ctime = GetTime()
	update_time = update_time + (ctime - update_this_start_time)

	local elapsed = ctime - update_init_time
	if elapsed >= 5 then
		debug_fs:SetText(string.format("%0.2f%%", update_time/elapsed*100))

		update_time = 0
		update_init_time = nil
	end
end
-- debug end

function GridStatusPriestAoe:OnUpdate(elapsed)
	update_timer = update_timer + elapsed

	-- check cooldowns here to update faster
	if coh_prev_best_guid and settings_coh.enable and settings_coh.hidecd and (not settings_coh.enable_if_have or coh_targets) and not IsCohReady() then
		self:ClearCohStatus()
	end

	if casc_prev_best_guid and settings_casc.enable and settings_casc.hidecd and (not settings_casc.enable_if_have or have_casc) and not IsCascReady() then
		self:ClearCascStatus()
	end

	if update_timer >= settings.cycle_time then
		update_timer = 0
		if settings.debug then start_timing() end
		GridStatusPriestAoe:RefreshAll()
		if settings.debug then end_timing() end
	end
end

local function DistanceSq(x1, y1, x2, y2)
	local xx = x2 - x1
	local yy = y2 - y1

	return xx*xx + yy*yy
end

local function Distance(x1, y1, x2, y2)
	local xx = x2 - x1
	local yy = y2 - y1

	return math_sqrt(xx*xx + yy*yy)
end

local function IsValidTarget(unitid)
	return not UnitIsDeadOrGhost(unitid) and
				UnitIsConnected(unitid) and
				UnitIsVisible(unitid) and
				not (UnitIsCharmed(unitid) or UnitIsEnemy("player", unitid))
end

local function IsUnitInRange(unitid)
	if UnitIsUnit(unitid, "player") then
		return true
	else
		return UnitInRange(unitid)
	end
end

function GridStatusPriestAoe:RefreshMapData()
	-- check player position
	refresh_state_player_x, refresh_state_player_y = UnitPosition("player")

	-- continue only if map supported
	if refresh_state_player_x and refresh_state_player_y then
		return true
	end

	self:Debug("UnitPosition failed")
	return false
end

function GridStatusPriestAoe:RefreshAll()
	if not self:RefreshMapData() then
		if settings_coh.enable then
			self:ClearCohStatus()
		end

		if settings_poh.enable then
			self:ClearPohStatus()
		end

		if settings_casc.enable then
			self:ClearCascStatus()
		end

		--[[
		if settings_hn.enable then
			self:ClearHnStatus()
		end
		]]
	else
		UpdateBonusCoefs()

		-- cache player data
		wipe(refresh_state_player_data)
		for group = 1, num_groups do
			for _, unitid in pairs(roster[group]) do
				local cx, cy = UnitPosition(unitid)
				if (cx ~= 0 or cy ~= 0) and IsValidTarget(unitid) then
					local health = UnitHealth(unitid)
					local health_max = UnitHealthMax(unitid)
					local cguid = UnitGUID(unitid)
					local percent = health / health_max

					refresh_state_player_data[unitid] = {
						guid = cguid,
						x = cx,
						y = cy,
						deficit = health_max - health,
						percent = percent,
						incoming = self:GetIncomingHeals(unitid),
						healmod = self:GetHealModifier(cguid) * (percent <= 0.5 and test_of_faith_mul or 1)
					}
				end
			end
		end
		
		-- coh
		if settings_coh.enable and
			(not settings_coh.hidecd or IsCohReady()) and
			(not settings_coh.enable_if_have or coh_targets) then
			self:RefreshCoH()
		end

		-- poh
		if settings_poh.enable then
			self:RefreshPoH()
		end

		-- cascade
		if settings_casc.enable and
			(not settings_casc.hidecd or IsCascReady()) and
			(not settings_casc.enable_if_have or have_casc) then
			self:RefreshCasc()
		end

		--[[ Holy Nova
		if settings_hn.enable then
			self:RefreshHN()
		end
		]]
	end
end

-- Circle of Healing
local function CohSortTargets(x, y)
	return x.percent < y.percent
end

local coh_temp_targets = {}
function GridStatusPriestAoe:RefreshCoH()
	local coh_avg = GetCohAvg()
	local threshold = settings_coh.threshold_targets * coh_avg

	local coh_best_guid
	local coh_best_heal = -1
	local coh_best_dists
	local coh_best_targets

	for unitid, p1 in pairs(refresh_state_player_data) do
		if not settings_coh.range or IsUnitInRange(unitid) then
			wipe(coh_temp_targets)
			for tunitid, p2 in pairs(refresh_state_player_data) do
				local dist = p1 == p2 and 0 or DistanceSq(p1.x, p1.y, p2.x, p2.y)
				if dist <= coh_testrange_sq and p2.deficit > 0 then
					local amount = math_min(p2.deficit, coh_avg * p2.healmod)
					if amount > 0 then
						tinsert(coh_temp_targets, { amount = amount, percent = p2.percent, dist = dist })
					end
				end
			end

			-- order candidate targets, get best
			if #coh_temp_targets > (coh_targets or 5) then
				tsort(coh_temp_targets, CohSortTargets)
			end

			local dists = 0
			local hsum = 0
			local this_targets = math_min(coh_targets or 5, #coh_temp_targets)
			for ti = 1, this_targets do
				hsum = hsum + coh_temp_targets[ti].amount
				dists = math_max(dists, coh_temp_targets[ti].dist)
			end
			-- dists = dists / math_min(coh_targets or 5, #targets)

			-- select best, buy try to keep same target
			if hsum > coh_best_heal or
					(hsum == coh_best_heal and (
						p1.guid == coh_prev_best_guid or
						(coh_best_guid ~= coh_prev_best_guid and dists < coh_best_dists)
						)) then
				coh_best_guid = p1.guid
				coh_best_heal = hsum
				coh_best_dists = dists
				coh_best_targets = this_targets
			end
		end
	end

	-- send status to best
	if coh_best_guid ~= coh_prev_best_guid or coh_best_heal < threshold then
		self:ClearCohStatus()
	end

	if coh_best_guid and coh_best_heal >= threshold then
		if coh_prev_best_guid ~= coh_best_guid or coh_prev_best_heal ~= coh_best_heal then
			self.core:SendStatusGained(	coh_best_guid,
										"alert_priestaoe_coh",
										settings_coh.priority,
										nil,
										settings_coh.color,
										tostring(math_floor(coh_best_heal + 0.5)),
										1,
										coh_best_targets,
										ICON_COH)
		end
		
		if Emphasize and settings_coh.emphasize and coh_prev_best_guid ~= coh_best_guid then
			Emphasize:EmphasizeUnit(coh_best_guid, "priestaoe_coh", settings_coh.emphasize_priority, settings_coh.emphasize_color)
		end
		
		coh_prev_best_guid = coh_best_guid
		coh_prev_best_heal = coh_best_heal
	end
end

-- Prayer of Healing
function GridStatusPriestAoe:RefreshPoH()
	local poh_avg = GetPohAvg()
	local threshold = settings_poh.threshold_targets * poh_avg

	self:ClearPohGridStatus()

	local best_group_guid = nil
	local best_group_amount = nil


	-- check all
	for group = 1, num_groups do
		local group_best_uid = nil
		local group_best_amount = -1
		local group_best_targets = -1
		local group_best_pdata = nil

		for _, unitid in pairs(roster[group]) do
			if not settings_poh.range or IsUnitInRange(unitid) then
				local amount = 0
				local targets = 0

				local p1 = refresh_state_player_data[unitid]
				if p1 then
					for _, tunitid in pairs(roster[group]) do
						local p2 = refresh_state_player_data[tunitid]
						if p2 then
							local dist = DistanceSq(p1.x, p1.y, p2.x, p2.y)
							if dist <= poh_testrange_sq then
								amount = amount + math_min(math_max(0, p2.deficit - p2.incoming), poh_avg * p2.healmod)
								targets = targets + 1
							end
						end
					end

					if amount > group_best_amount or (amount == group_best_amount and targets > group_best_targets) then
						group_best_amount = amount
						group_best_uid = unitid
						group_best_targets = targets
						group_best_pdata = p1
					end
				end
			end
		end
		
		if group_best_uid and (not best_group_guid or group_best_amount > best_group_amount) then
			best_group_guid = group_best_pdata.guid
			best_group_amount = group_best_amount
		end

		if group_best_uid and ((settings_poh.always_best and group_best_targets >= 3) or group_best_amount >= threshold) then
			local text = (group_best_amount >= threshold) and math_floor(group_best_amount + 0.5) or group_best_targets
			self.core:SendStatusGained(	group_best_pdata.guid,
										"alert_priestaoe_poh",
										settings_poh.priority,
										nil,
										settings_poh.color,
										tostring(text),
										1,
										group_best_targets,
										ICON_POH)
		end
	end
	
	if Emphasize and settings_poh.emphasize then
		if best_group_amount and best_group_amount > threshold then
			if poh_prev_best ~= best_group_guid then
				self:ClearPohEmphasizeStatus()
				poh_prev_best = best_group_guid
				Emphasize:EmphasizeUnit(best_group_guid, "priestaoe_poh", settings_poh.emphasize_priority, settings_poh.emphasize_color)
			end
		else
			self:ClearPohEmphasizeStatus()
		end
	end
end

local function DistSortTargets(x, y)
	return y[1] < x[1]
end

local function CascadeHeal(pdata, prev_heal, dist)
	local max_heal = GetCascAvg(dist)
	local cur_deficit = math_max(pdata.deficit - prev_heal, 0)
	local heal = math_min(cur_deficit, max_heal)

	return heal
end

local function CascadeJumps(from, touched)
	if not from then
		return nil, nil, 0
	end

	-- create distances table if needed
	if not from.dists then
		local fx, fy = from.x, from.y
		from.dists = {}
		for unitid, p2 in pairs(refresh_state_player_data) do
			if p2.guid ~= from.guid then
				local dist_sq = DistanceSq(fx, fy, p2.x, p2.y)
				if dist_sq <= CASC_JUMP_RANGE_SQ then
					local dd = { dist_sq, p2 }
					tinsert(from.dists, dd)
				end
			end
		end
		tsort(from.dists, DistSortTargets)
	end

	-- test possible targets
	local dists = from.dists
	local num_targets = #dists
	local found_targets = {}
	local heal = 0
	for i = 1, num_targets do
		local target = dists[i]
		local target_dist_sq = target[1]
		local target_data = target[2]

		local prev_heal = touched[target_data] or 0
		local this_heal = CascadeHeal(target_data, prev_heal, math_sqrt(target_dist_sq))
		touched[target_data] = prev_heal + this_heal
		heal = heal + this_heal

		tinsert(found_targets, target_data)
		if #found_targets == 2 then
			break
		end
	end

	return found_targets[1], found_targets[2], heal
end

function GridStatusPriestAoe:RefreshCasc()
	local casc_avg = GetCascAvg(40)
	local threshold = settings_casc.threshold_targets * casc_avg

	local best_guid
	local best_heal = 0
	local px, py = refresh_state_player_x, refresh_state_player_y	
	local touched = {}
	local best_jumps

	for unitid, p2 in pairs(refresh_state_player_data) do
		if IsUnitInRange(unitid) then
			local heal

			wipe(touched)
			heal = CascadeHeal(p2, 0, Distance(px, py, p2.x, p2.y))
			touched[p2] = heal

			-- 2 jumps
			local t1, t2, p2heal = CascadeJumps(p2, touched)

			-- 2*2 jumps
			local t3, t4, p3heal = CascadeJumps(t1, touched)
			local t5, t6, p4heal = CascadeJumps(t2, touched)

			-- 2*4 jumps
			local t7, t8, p5heal = CascadeJumps(t3, touched)
			local t9, t10, p6heal = CascadeJumps(t4, touched)
			local t11, t12, p7heal = CascadeJumps(t5, touched)
			local t13, t14, p8heal = CascadeJumps(t6, touched)

			-- 2*8 jumps
			local t15, t16, p8heal = CascadeJumps(t7, touched)
			local t17, t18, p9heal = CascadeJumps(t8, touched)
			local t19, t20, p10heal = CascadeJumps(t9, touched)
			local t21, t22, p11heal = CascadeJumps(t10, touched)
			local t23, t24, p12heal = CascadeJumps(t11, touched)
			local t25, t26, p13heal = CascadeJumps(t12, touched)
			local t27, t28, p14heal = CascadeJumps(t13, touched)
			local t29, t30, p15heal = CascadeJumps(t14, touched)

			heal = heal + p2heal + p3heal + p4heal + p5heal + p6heal + p7heal + p8heal +
				p9heal + p10heal + p11heal + p12heal + p13heal + p14heal + p15heal

			if heal > 0 and heal > best_heal then
				best_guid = p2.guid
				best_heal = heal
				-- best_jumps = { p2, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14 }
			end
		end
	end
	---------------------

	self:ClearCascStatus()

	if best_guid and best_heal >= threshold then
		self.core:SendStatusGained(	best_guid,
									"alert_priestaoe_casc",
									settings_casc.priority,
									nil,
									settings_casc.color,
									tostring(math_floor(best_heal + 0.5)),
									1,
									nil,
									ICON_CASC)

		if Emphasize and settings_casc.emphasize and casc_prev_best_guid ~= best_guid then
			Emphasize:EmphasizeUnit(best_guid, "priestaoe_casc", settings_casc.emphasize_priority, settings_casc.emphasize_color)
		end
				
		casc_prev_best_guid = best_guid
		casc_prev_best_heal = best_heal
	else
		casc_prev_best_guid = nil
	end
end


--[[ Holy Nova
local function HnSortTargets(x, y)
	return x.dist < y.dist
end

local inrange = {}
function GridStatusPriestAoe:RefreshHN()
	self:ClearHnStatus()

	if settings.first_groups_only and player_group > num_groups then
		return
	end

	local hn_avg = GetHnAvg()

	wipe(inrange)

	local amount = 0

	for unitid, p2 in pairs(refresh_state_player_data) do
		if p2.guid == player_guid then
			amount = math_min(p2.deficit, hn_avg * p2.healmod)
		else
			local dist = DistanceSq(refresh_state_player_x, refresh_state_player_y, p2.x, p2.y)
			if dist <= hn_testrange_sq then
				tinsert(inrange, { player = p2, dist = dist })
			end
		end
	end

	-- sort targets by distance and select the 4 closest to the player
	if #inrange > 4 then
		tsort(inrange, HnSortTargets)
	end

	for ti = 1, math_min(4, #inrange) do
		amount = amount + math_min(inrange[ti].player.deficit, hn_avg * inrange[ti].player.healmod)
	end

	if amount >= settings_hn.threshold then
		self.core:SendStatusGained(	player_guid,
									"alert_priestaoe_hn",
									settings_hn.priority,
									nil,
									settings_hn.color,
									tostring(math_floor(amount + 0.5)),
									1,
									nil,
									ICON_HN)
		if settings_hn.emphasize and Emphasize then
			Emphasize:EmphasizeUnit(player_guid, "priestaoe_hn", settings_hn.color)
		end
	end
end
]]

function GridStatusPriestAoe:HasGlyph(gid)
	for i = 1, 9 do
		local _, _, _, id = GetGlyphSocketInfo(i)
		if id and gid == id then
			return true
		end
	end
	return false
end

function GridStatusPriestAoe:GetSetAmount(set)
	local amount = 0
	for _, itemID in ipairs(set) do
		if IsEquippedItem(itemID) then
			amount = amount + 1
		end
	end
	return amount
end

function GridStatusPriestAoe:UnitHasPlayerAura(unitid, aura)
	local name, _, _, _, _, _, expires, caster = UnitBuff(unitid, aura)
	return name and caster == "player" and expires
end

function GridStatusPriestAoe:GetHealModifier(guid)
	return 1
end

function GridStatusPriestAoe:GetIncomingHeals(unitid)
	if settings_poh.enable and settings_poh.incoming_heals then
		return (UnitGetIncomingHeals(unitid) or 0) - (UnitGetIncomingHeals(unitid, "player") or 0)
	end
	return 0
end
